;; Generated package description from oauth2.el  -*- no-byte-compile: t; lexical-binding:t -*-
(define-package "oauth2" "0.19.1snapshot0.20260526.1" "OAuth 2.0 Authorization Protocol" '((emacs "27.1")) :commit "9697975d6c2d721dfe9733f103eb4e5a6f491ad9" :authors '(("Julien Danjou" . "julien@danjou.info")) :maintainer '(("Xiyue Deng" . "manphiz@gmail.com") (nil . "emacs-devel@gnu.org")) :keywords '("comm") :url "https://elpa.gnu.org/packages/oauth2.html")
